from linerate.units import Meter, Unitless, WattPerMeter, WattPerSquareMeter


def compute_solar_heating(
    absorptivity: Unitless,
    global_radiation_intensity: WattPerSquareMeter,
    conductor_diameter: Meter,
) -> WattPerMeter:
    r"""Compute the solar heating experienced by the conductor.

    Equation (8) on page 18 of :cite:p:`cigre601` and (11) on page 4 in :cite:p:`cigre207`.

    Parameters
    ----------
    absorptivity:
        :math:`\alpha_s`. Material constant. According to :cite:p:`cigre601`, it starts at
        approximately 0.2 for new cables and reaches a constant value of approximately 0.9
        after about one year.
    global_radiation_intensity:
        :math:`I_T~\left[\text{W}~\text{m}^{-2}\right]`.The global radiation intensity.
    conductor_diameter:
        :math:`D~\left[\text{m}\right]`. Outer diameter of the conductor.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`P_S~\left[\text{W}~\text{m}^{-1}\right]`. The solar heating of the conductor
    """
    alpha_s = absorptivity
    I_T = global_radiation_intensity
    D = conductor_diameter

    return alpha_s * I_T * D



import warnings

import numpy as np
from numba import vectorize

from ...units import (
    Celsius,
    KilogramPerCubeMeter,
    KilogramPerMeterPerSecond,
    Meter,
    Radian,
    SquareMeterPerSecond,
    Unitless,
    WattPerMeter,
    WattPerMeterPerKelvin,
)

# Physical quantities
#####################


def compute_temperature_gradient(
    total_heat_gain: WattPerMeter,
    conductor_thermal_conductivity: WattPerMeterPerKelvin,
    core_diameter: Meter,
    conductor_diameter: Meter,
) -> Celsius:
    r"""Compute the difference between the core and surface temperature.

    Equation (15) & (16) on page 22 of :cite:p:`cigre601`.

    Parameters
    ----------
    total_heat_gain:
        :math:`P_T = I^2 R~\left[\text{W}~\text{m}^{-1}\right]`. The Joule heating of the
        conductor (see p. 81 of :cite:p:`cigre601`).
    conductor_thermal_conductivity:
        :math:`\lambda \left[\text{W}~\text{m}^{-1}~\text{K}^{-1}\right]`. The effective
        conductor thermal conductivity. It is usually between :math:`0.5` and
        :math:`7~W~m^{-1}~K^{-1}`. Recommended values are
        :math:`0.7~\text{W}~\text{m}^{-1}~\text{K}^{-1}` for conductors with no tension on the
        aluminium strands and :math:`1.5~\text{W}~\text{m}^{-1}~\text{K}^{-1}` for conductors
        with aluminium strands under a tension of at least 40 N :cite:p:`cigre601`.
    core_diameter:
        :math:`D_1~\left[\text{m}\right]`. Diameter of the steel core of the conductor.
    conductor_diameter:
        :math:`D~\left[\text{m}\right]`. Outer diameter of the conductor.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`T_c - T_s~\left[^\circ\text{C}\right]`. The difference between the core and
        surface temperature.
    """
    lambda_ = conductor_thermal_conductivity
    D_1 = core_diameter
    D = conductor_diameter
    P_T = total_heat_gain
    pi = np.pi

    tmp = P_T / (2 * pi * lambda_)

    if D_1 == 0:  # TODO: Maybe lower tolerance?
        return 0.5 * tmp
    else:
        D_1_sq = D_1**2
        delta_D_sq = D**2 - D_1_sq
        return tmp * (0.5 - (D_1_sq / delta_D_sq) * np.log(D / D_1))


def compute_thermal_conductivity_of_air(film_temperature: Celsius) -> WattPerMeterPerKelvin:
    r"""Approximation of the thermal conductivity of air up to :math:`300 ^\circ\text{C}`.

    Equation (18) on page 24 of :cite:p:`cigre601`.

    Compared with table values from textbook, which showed a good approximation.

    Parameters
    ----------
    film_temperature:
        :math:`T_f = 0.5 (T_s + T_a)~\left[^\circ\text{C}\right]`. The temperature of the
        thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor sufrace temperature.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\lambda_f~\left[\text{W}~\text{m}^{-1}~\text{K}^{-1}\right]`. The thermal
        conductivity of air at the given temperature.
    """
    T_f = film_temperature
    return 2.368e-2 + 7.23e-5 * T_f - 2.763e-8 * (T_f**2)


def compute_air_density(
    film_temperature: Celsius, height_above_sea_level: Meter
) -> KilogramPerCubeMeter:
    r"""Approximation of the density of air at a given temperature and altitude.

    Equation (20) on page 25 of :cite:p:`cigre601`.

    Parameters
    ----------
    film_temperature:
        :math:`T_f = 0.5 (T_s + T_a)~\left[^\circ\text{C}\right]`. The temperature of the
        thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor sufrace temperature.
    height_above_sea_level:
        :math:`y~\left[\text{m}\right]`. The conductor's altitude.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\gamma~\left[\text{kg}~\text{m}^{-3}\right]`. The mass density of air.
    """
    T_f = film_temperature
    y = height_above_sea_level
    return (1.293 - 1.525e-4 * y + 6.379e-9 * (y**2)) / (1 + 0.00367 * T_f)


def compute_dynamic_viscosity_of_air(film_temperature: Celsius) -> KilogramPerMeterPerSecond:
    r"""Approximation of the dynamic viscosity of air at a given temperature.

    Equation (19) on page 25 of :cite:p:`cigre601`.

    Parameters
    ----------
    film_temperature:
        :math:`T_f = 0.5 (T_s + T_a)~\left[^\circ\text{C}\right]`. The temperature of the
        thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor sufrace temperature.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\mu_f~\left[\text{kg}~\text{m}^{-1}~\text{s}^{-1}\right]`. The dynamic viscosity
        of air.
    """
    T_f = film_temperature
    return 17.239e-6 + 4.635e-8 * T_f - 2.03e-11 * (T_f**2)


def compute_kinematic_viscosity_of_air(
    dynamic_viscosity_of_air: KilogramPerMeterPerSecond, air_density: KilogramPerCubeMeter
) -> SquareMeterPerSecond:
    r"""Compute the kinematic viscosity of air.

    Definition in text on page 25 of :cite:p:`cigre601`.

    Parameters
    ----------
    dynamic_viscosity_of_air:
        :math:`\mu_f~\left[\text{kg}~\text{m}^{-1}~\text{s}^{-1}\right]`. The dynamic viscosity of
        air.
    air_density:
        :math:`\gamma~\left[\text{kg}~\text{m}^{-3}\right]`. The mass density of air.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\nu_f~\left[\text{m}^2~\text{s}^{-1}\right]`. The kinematic viscosity of air.
    """
    return dynamic_viscosity_of_air / air_density


## Nusselt number calculation
#############################


def _check_perpendicular_flow_nusseltnumber_out_of_bounds(reynolds_number, conductor_roughness):
    Re = reynolds_number
    Rs = conductor_roughness
    if np.any(Re < 0):
        raise ValueError("Reynolds number cannot be negative.")

    smooth_mask = np.logical_or(Rs == 0, np.isnan(Rs))
    smooth_out_of_bounds = Re > 200_000
    rough_out_of_bounds = np.logical_and(np.logical_not(smooth_mask), Re > 50_000)
    out_of_bounds = np.logical_or(smooth_out_of_bounds, rough_out_of_bounds)

    if np.any(out_of_bounds):
        warnings.warn("Reynolds number is out of bounds", stacklevel=5)


@vectorize(nopython=True)
def _compute_perpendicular_flow_nusseltnumber(
    reynolds_number: Unitless,
    conductor_roughness: Meter,
) -> Unitless:
    # TODO: Look at references for this table
    Re = reynolds_number
    Rs = conductor_roughness

    if Rs == 0 or np.isnan(Rs):
        if Re < 35:
            B, n = 0, 0
        elif Re < 5000:
            B, n = 0.583, 0.471
        elif Re < 50_000:
            B, n = 0.148, 0.633
        else:
            B, n = 0.0208, 0.814
    elif Rs <= 0.05:
        if Re < 100:
            B, n = 0, 0
        elif Re < 2650:
            B, n = 0.641, 0.471
        else:
            B, n = 0.178, 0.633
    else:
        if Re < 100:
            B, n = 0, 0
        elif Re < 2650:
            B, n = 0.641, 0.471
        else:
            B, n = 0.048, 0.800

    return B * Re**n  # type: ignore


def compute_perpendicular_flow_nusseltnumber(
    reynolds_number: Unitless,
    conductor_roughness: Meter,
) -> Unitless:
    r"""Compute the Nusselt number for perpendicular flow.

    Equation (21) and Table 4 on pages 25-26 of :cite:p:`cigre601`.

    The perpendicular flow nusselt number is denoted as :math:`\text{Nu}_{90}` in :cite:p:`cigre601`
    since the wind's angle of attack is :math:`90^\circ`.

    The Nusselt number is the ratio of conductive heat transfer to convective heat transfer.

    Parameters
    ----------
    reynolds_number:
        :math:`\text{Re}`. The Reynolds number.
    conductor_roughness:
        :math:`\text{Rs}`. The roughness number

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Nu}_{90}`. The perpendicular flow Nusselt number.
    """
    _check_perpendicular_flow_nusseltnumber_out_of_bounds(reynolds_number, conductor_roughness)
    return _compute_perpendicular_flow_nusseltnumber(
        reynolds_number,
        conductor_roughness,
    )


@vectorize(nopython=True)
def _correct_wind_direction_effect_on_nusselt_number(
    perpendicular_flow_nusselt_number: Unitless,
    angle_of_attack: Radian,
    conductor_roughness: Unitless,
) -> Unitless:
    delta = angle_of_attack
    Nu_90 = perpendicular_flow_nusselt_number
    Rs = conductor_roughness

    sin_delta = np.sin(delta)

    if Rs == 0 or np.isnan(Rs):
        sin_delta_sq = sin_delta**2
        cos_delta_sq = 1 - sin_delta_sq

        correction_factor = (sin_delta_sq + cos_delta_sq * 0.0169) ** 0.225
    else:
        if delta <= np.radians(24):
            correction_factor = 0.42 + 0.68 * (sin_delta**1.08)
        else:
            correction_factor = 0.42 + 0.58 * (sin_delta**0.90)

    return correction_factor * Nu_90


def correct_wind_direction_effect_on_nusselt_number(
    perpendicular_flow_nusselt_number: Unitless,
    angle_of_attack: Radian,
    conductor_roughness: Unitless,
) -> Unitless:
    r"""Correct the Nusselt number for the wind's angle-of-attack.

    Equation (21) and Table 4 on pages 25-26 of :cite:p:`cigre601`.

    The perpendicular flow nusselt number is denoted as :math:`\text{Nu}_\delta` in
    :cite:p:`cigre601` since the wind's angle of attack is :math:`\delta`.

    Parameters
    ----------
    perpendicular_flow_nusselt_number:
        :math:`\text{Nu}_{90}`. The perpendicular flow Nusselt number.
    angle_of_attack:
        :math:`\delta~\left[\text{radian}\right]`. The wind angle-of-attack.
    conductor_roughness:
        :math:`\text{Rs}`. The roughness number

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Nu}_\delta`. The Nusselt number for the given wind angle-of-attack.
    """
    return _correct_wind_direction_effect_on_nusselt_number(
        perpendicular_flow_nusselt_number,
        angle_of_attack,
        conductor_roughness,
    )


## Natural convection computations (no wind):
#############################################


def _check_horizontal_natural_nusselt_number(
    grashof_number: Unitless, prandtl_number: Unitless
) -> None:
    GrPr = grashof_number * prandtl_number
    if np.any(GrPr < 0):
        raise ValueError("GrPr cannot be negative.")
    elif np.any(GrPr > 1e12):
        raise ValueError("GrPr out of bounds: Must be < 10^12.")


@vectorize(nopython=True)
def _compute_horizontal_natural_nusselt_number(
    grashof_number: Unitless,
    prandtl_number: Unitless,
) -> Unitless:
    GrPr = grashof_number * prandtl_number

    if GrPr < 1e-1:
        return 0
    elif GrPr < 1e2:
        return 1.020 * GrPr**0.148
    elif GrPr < 1e4:
        return 0.850 * GrPr**0.188
    elif GrPr < 1e7:
        return 0.480 * GrPr**0.250
    else:
        return 0.125 * GrPr**0.333


def compute_horizontal_natural_nusselt_number(
    grashof_number: Unitless,
    prandtl_number: Unitless,
) -> Unitless:
    r"""The Nusselt number for natural (passive) convection on a horizontal conductor.

    Equation (23) and Table 5 on pages 27-28 of :cite:p:`cigre601`.

    The natural convection Nusselt number is denoted by both :math:`\text{Nu}_\text{nat}`
    and :math:`\text{Nu}_0` (due to the conductor declination being :math:`0^\circ`)
    in :cite:p:`cigre601`.

    The coefficient table is modified slightly so coefficients with
    :math:`\text{Gr}\text{Pr} < 0.1` leads to :math:`\text{Nu} = 0`.

    Parameters
    ----------
    grashof_number:
        :math:`\text{Gr}`. The Grashof number.
    prandtl_number:
        :math:`\text{Pr}`. The Prandtl number.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Nu}_0`. The natural convection nusselt number assuming horizontal conductor.
    """
    _check_horizontal_natural_nusselt_number(grashof_number, prandtl_number)
    return _compute_horizontal_natural_nusselt_number(
        grashof_number,
        prandtl_number,
    )


def _check_conductor_inclination(
    conductor_inclination: Radian,
    conductor_roughness: Unitless,
) -> None:
    beta = np.degrees(conductor_inclination)
    Rs = conductor_roughness

    smooth_mask = np.logical_or(Rs == 0, np.isnan(Rs))
    smooth_out_of_bounds = np.logical_and(smooth_mask, beta > 60)
    rough_out_of_bounds = np.logical_and(np.logical_not(smooth_mask), beta > 80)
    out_of_bounds = np.logical_or(smooth_out_of_bounds, rough_out_of_bounds)
    if np.any(out_of_bounds):
        raise ValueError(
            "Inclination must be < 60° for smooth conductors and 80° for stranded conductors"
        )


@vectorize(nopython=True)
def _correct_natural_nusselt_number_inclination(
    horizontal_natural_nusselt_number: Unitless,
    conductor_inclination: Radian,
    conductor_roughness: Unitless,
) -> Unitless:
    beta = np.degrees(conductor_inclination)
    Nu_nat = horizontal_natural_nusselt_number
    Rs = conductor_roughness

    if Rs == 0 or np.isnan(Rs):
        return Nu_nat * (1 - 1.58e-4 * beta**1.5)
    else:
        return Nu_nat * (1 - 1.76e-6 * beta**2.5)


def correct_natural_nusselt_number_inclination(
    horizontal_natural_nusselt_number: Unitless,
    conductor_inclination: Radian,
    conductor_roughness: Unitless,
) -> Unitless:
    r"""Correct the natural Nusselt number for the effect of the span inclination.

    Equation (24) on page 28 of :cite:p:`cigre601`.

    Parameters
    ----------
    horizontal_natural_nusselt_number:
        :math:`\text{Nu}_0`. The natural convection nusselt number assuming horizontal conductor.
    conductor_inclination:
        :math:`\beta~\left[\text{radian}\right]`. The inclination angle of the conductor. The
        inclination can be computed as
        :math:`\beta = \text{arctan2}\left(\left|y_1 - y_0\right|, L\right)`, where :math:`y_0`
        and :math:`y_1` are the altitude of the span endpoints (towers) and :math:`L` is the
        length of the span .
    conductor_roughness:
        :math:`Rs`. The roughness number.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Nu}_\beta`. The natural convection nusselt number where the conductor
        inclination is taken into account.
    """
    _check_conductor_inclination(conductor_inclination, conductor_roughness)
    return _correct_natural_nusselt_number_inclination(
        horizontal_natural_nusselt_number,
        conductor_inclination,
        conductor_roughness,
    )


def compute_nusselt_number(
    forced_convection_nusselt_number: Unitless,
    natural_nusselt_number: Unitless,
) -> Unitless:
    r"""Compute the nusselt number.

    Described in the text on p. 28 of :cite:p:`cigre601`.

    Parameters
    ----------
    forced_convection_nusselt_number:
        :math:`\text{Nu}_\delta`. The Nusselt number for the given wind angle-of-attack.
    natural_nusselt_number:
        :math:`\text{Nu}_\delta`. The natural convection nusselt number where the conductor
        inclination is taken into account.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`Nu`. The nusselt number.
    """
    return np.maximum(forced_convection_nusselt_number, natural_nusselt_number)



from dataclasses import dataclass
from functools import cached_property
from typing import Optional

import numpy as np
import pygeodesy

from .units import (
    Celsius,
    Degrees,
    Meter,
    MeterPerSecond,
    OhmPerMeter,
    Radian,
    SquareMeter,
    SquareMeterPerAmpere,
    Unitless,
    WattPerMeterPerKelvin,
)

__all__ = ["Conductor", "Weather", "Tower", "Span"]


@dataclass(frozen=True)
class Conductor:
    """Container for conductor parameters."""

    #: :math:`D_1~\left[\text{m}\right]`. Diameter of the steel core of the conductor.
    core_diameter: Meter

    #: :math:`D~\left[\text{m}\right]`. Outer diameter of the conductor.
    conductor_diameter: Meter

    #: :math:`d~\left[\text{m}\right]`. The diameter of the strands in the outer layer of
    #: the conductor.
    outer_layer_strand_diameter: Meter

    #: :math:`\epsilon_s`. The emmisivity of the conductor.
    emissivity: Unitless

    #: :math:`\alpha_s`. Material constant. According to :cite:p:`cigre601`, it starts at
    #: approximately 0.2 for new cables and reaches a constant value of approximately 0.9
    #: after about one year.
    solar_absorptivity: Unitless

    #: :math:`T_1~\left[^\circ C\right]`. The first temperature with known resistance
    temperature1: Celsius
    #: :math:`T_2~\left[^\circ C\right]`. The second temperature with known resistance
    temperature2: Celsius
    #: :math:`R_1~\left[\Omega \text{m}^{-1}\right]`. (AC-)resistance at temperature :math:`T_1`
    resistance_at_temperature1: OhmPerMeter
    #: :math:`R_2~\left[\Omega \text{m}^{-1}\right]`. (AC-)resistance at temperature :math:`T_2`
    resistance_at_temperature2: OhmPerMeter

    #: :math:`A_{\text{Al}}~\left[\text{m}^2\right]`. The cross sectional area of the aluminium
    #: strands in the conductor. Used for correcting for magnetic core effects in ACSR conductors.
    aluminium_cross_section_area: SquareMeter
    #: :math:`b`. The constant magnetic effect, most likely equal to 1. If ``None``, then no
    #: correction is used (used for non-ACSR cables).
    constant_magnetic_effect: Unitless
    #: :math:`m`. The current density proportional magnetic effect. If ``None``, then it is assumed
    #: equal to 0.
    current_density_proportional_magnetic_effect: SquareMeterPerAmpere
    #: :math:`c_\text{max}`. Saturation point of the relative increase in conductor resistance due
    # to magnetic core effects.
    max_magnetic_core_relative_resistance_increase: Unitless

    #: :math:`\lambda \left[\text{W}~\text{m}^{-1}~\text{K}^{-1}\right]`. The effective
    #: conductor thermal conductivity. It is usually between :math:`0.5` and
    #: :math:`7~W~m^{-1}~K^{-1}`. Recommended values are
    #: :math:`0.7~\text{W}~\text{m}^{-1}~\text{K}^{-1}` for conductors with no tension on the
    #: aluminium strands and :math:`1.5~\text{W}~\text{m}^{-1}~\text{K}^{-1}` for conductors
    #: with aluminium strands under a tension of at least 40 N :cite:p:`cigre601`.
    thermal_conductivity: Optional[WattPerMeterPerKelvin] = None


@dataclass(frozen=True)
class Tower:
    """Container for a tower (span end-point)."""

    #: :math:`\phi~\left[^\circ\right]`. The tower's longitude (east of the prime meridian).
    longitude: Degrees
    #: The tower's latitude (north-facing).
    latitude: Degrees
    #: :math:`y~\left[m\right]`. The tower's altitude.
    altitude: Meter


@dataclass(frozen=True)
class Span:
    """Container for a span.

    Note
    ----
    For more information about the albedo, see
    :py:func:`linerate.equations.solar_heating.compute_global_radiation_intensity` for a table of
    albedo values for different ground types.
    """

    #: Container for the conductor metadata
    conductor: Conductor
    #: Container for the metadata of the first tower of the span
    start_tower: Tower
    #: Container for the metadata of the second tower of the span
    end_tower: Tower

    #: :math:`F`.  The ground albedo.
    ground_albedo: Unitless

    #: Number of conductors in the span. 1 for simplex, 2 for duplex and 3 for triplex.
    num_conductors: Unitless

    @cached_property
    def latitude(self) -> Degrees:
        r""":math:`\phi~\left[^\circ\right]`. The latitude of the span midpoint."""
        return 0.5 * (self.start_tower.latitude + self.end_tower.latitude)

    @cached_property
    def longitude(self) -> Degrees:
        r""":math:`\left[^\circ\right]`. The longitude of the span midpoint."""
        return 0.5 * (self.start_tower.longitude + self.end_tower.longitude)

    @cached_property
    def inclination(self) -> Radian:
        r""":math:`\beta~\left[\text{radian}\right]`. The inclination.

        The inclination is computed from the difference in span altitude and the span length.
        """
        delta_y = np.abs(self.end_tower.altitude - self.start_tower.altitude)
        return np.arctan2(delta_y, self.span_length)

    @cached_property
    def conductor_azimuth(self) -> Radian:
        r""":math:`\gamma_c~\left[\text{radian}\right]`. Angle (east of north) the span is facing"""
        bearing = np.vectorize(pygeodesy.formy.bearing)
        return np.radians(  # type: ignore
            bearing(
                lat1=self.start_tower.latitude,
                lon1=self.start_tower.longitude,
                lat2=self.end_tower.latitude,
                lon2=self.end_tower.longitude,
            )
        )

    @cached_property
    def span_length(self) -> Meter:
        r""":math:`\left[\text{km}\right]`. The span length.

        The span length is computed with the haversine formula (assuming spherical earth).
        """
        haversine = np.vectorize(pygeodesy.formy.haversine)
        return haversine(  # type: ignore
            lat1=self.start_tower.latitude,
            lon1=self.start_tower.longitude,
            lat2=self.end_tower.latitude,
            lon2=self.end_tower.longitude,
        )

    @cached_property
    def conductor_altitude(self) -> Meter:
        r""":math:`y~\left[\text{m}\right]`. The span altitude.

        The altitude is computes as the average of the tower altitudes.
        """
        return 0.5 * (self.start_tower.altitude + self.end_tower.altitude)


@dataclass()
class Weather:
    #: :math:`T_a~\left[^\circ C\right]`. The ambient air temperature.
    air_temperature: Celsius
    #: :math:`\delta~\left[\text{radian}\right]`. Wind direction east of north.
    wind_direction: Radian
    #: :math:`v~\left[\text{m}~\text{s}^{-1}\right]`. Wind velocity
    wind_speed: MeterPerSecond
    #: :math:`N_s`. The clearness ratio (or clearness number in
    #: :cite:p:`sharma1965interrelationships,cigre207`).
    clearness_ratio: Unitless = 1



from linerate.equations import (
    cigre207,
    convective_cooling,
    dimensionless,
    math,
    solar_angles,
    solar_heating,
)
from linerate.equations.math import switch_cos_sin
from linerate.models.thermal_model import ThermalModel, _copy_method_docstring
from linerate.types import Span, Weather
from linerate.units import Ampere, Celsius, Date, OhmPerMeter, WattPerMeter


class Cigre207(ThermalModel):
    def __init__(
        self,
        span: Span,
        weather: Weather,
        time: Date,
        include_diffuse_radiation: bool = True,
        direct_radiation_factor: float = 1.0,
    ):
        super().__init__(span, weather)
        self.time = time
        self.include_diffuse_radiation = include_diffuse_radiation
        self.direct_radiation_factor = direct_radiation_factor

    @_copy_method_docstring(ThermalModel)
    def compute_joule_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_joule_heating(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_solar_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        alpha_s = self.span.conductor.solar_absorptivity
        phi = self.span.latitude
        gamma_c = self.span.conductor_azimuth
        y = self.span.conductor_altitude
        D = self.span.conductor.conductor_diameter

        omega = solar_angles.compute_hour_angle_relative_to_noon(self.time, self.span.longitude)
        delta = solar_angles.compute_solar_declination(self.time)
        sin_H_s = solar_angles.compute_sin_solar_altitude(phi, delta, omega)
        chi = solar_angles.compute_solar_azimuth_variable(phi, delta, omega)
        C = solar_angles.compute_solar_azimuth_constant(chi, omega)
        gamma_s = solar_angles.compute_solar_azimuth(C, chi)  # Z_c in IEEE
        cos_eta = solar_angles.compute_cos_solar_effective_incidence_angle(
            sin_H_s, gamma_s, gamma_c
        )
        sin_eta = switch_cos_sin(cos_eta)

        I_B = self.direct_radiation_factor * cigre207.solar_heating.compute_direct_solar_radiation(
            sin_H_s, y
        )
        if self.include_diffuse_radiation:
            I_d = cigre207.solar_heating.compute_diffuse_sky_radiation(I_B, sin_H_s)
            F = self.span.ground_albedo
        else:
            I_d = 0
            F = 0
        I_T = cigre207.solar_heating.compute_global_radiation_intensity(
            I_B, I_d, F, sin_eta, sin_H_s
        )
        return solar_heating.compute_solar_heating(
            alpha_s,
            I_T,
            D,
        )

    @_copy_method_docstring(ThermalModel)
    def compute_convective_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        D = self.span.conductor.conductor_diameter
        d = self.span.conductor.outer_layer_strand_diameter
        V = self.weather.wind_speed
        y = self.span.conductor_altitude
        T_a = self.weather.air_temperature
        T_c = conductor_temperature
        T_f = 0.5 * (T_c + T_a)

        # Compute physical quantities
        lambda_f = cigre207.convective_cooling.compute_thermal_conductivity_of_air(T_f)
        nu_f = cigre207.convective_cooling.compute_kinematic_viscosity_of_air(T_f)
        delta = math.compute_angle_of_attack(
            self.weather.wind_direction, self.span.conductor_azimuth
        )

        # Compute unitless quantities
        rho_r = cigre207.convective_cooling.compute_relative_air_density(y)
        Re = cigre207.convective_cooling.compute_reynolds_number(V, D, nu_f, rho_r)
        Gr = dimensionless.compute_grashof_number(D, T_c, T_a, nu_f)
        Pr = cigre207.convective_cooling.compute_prandtl_number(T_f)
        Rs = dimensionless.compute_conductor_roughness(D, d)

        # Compute nusselt numbers
        Nu_90 = cigre207.convective_cooling.compute_perpendicular_flow_nusseltnumber(
            reynolds_number=Re, conductor_roughness=Rs
        )
        Nu_delta = cigre207.convective_cooling.correct_wind_direction_effect_on_nusselt_number(
            Nu_90, delta
        )
        Nu_cor = cigre207.convective_cooling.compute_low_wind_speed_nusseltnumber(Nu_90)

        Nu_0 = cigre207.convective_cooling.compute_horizontal_natural_nusselt_number(Gr, Pr)

        Nu = cigre207.convective_cooling.compute_nusselt_number(
            forced_convection_nusselt_number=Nu_delta,
            natural_nusselt_number=Nu_0,
            low_wind_nusselt_number=Nu_cor,
            wind_speed=V,
        )

        return convective_cooling.compute_convective_cooling(
            surface_temperature=conductor_temperature,
            air_temperature=self.weather.air_temperature,
            nusselt_number=Nu,
            thermal_conductivity_of_air=lambda_f,
        )

    @_copy_method_docstring(ThermalModel)
    def compute_radiative_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_radiative_cooling(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_resistance(self, conductor_temperature: Celsius, current: Ampere) -> OhmPerMeter:
        return super().compute_resistance(
            conductor_temperature=conductor_temperature, current=current
        )



from textwrap import dedent

import numpy as np

from linerate.units import (
    Celsius,
    JoulePerKilogramPerKelvin,
    KilogramPerMeterPerSecond,
    Meter,
    MeterPerSecond,
    MeterPerSquareSecond,
    SquareMeterPerSecond,
    Unitless,
    WattPerMeterPerKelvin,
)


def compute_reynolds_number(
    wind_speed: MeterPerSecond,
    conductor_diameter: Meter,
    kinematic_viscosity_of_air: SquareMeterPerSecond,
) -> Unitless:
    r"""Compute the Reynolds number using the conductor diameter as characteristic length scale.

    Defined in the text on page 25 of :cite:p:`cigre601` and equation (2c)
    on page 10 in :cite:p:`ieee738`.

    The Reynolds number is a dimensionless quantity that can be used to assess if a stream is
    likely to be turbulent or not. It is given by

    .. math::

        \text{Re} = \frac{v L}{\nu},

    where :math:`v` is the flow velocity, :math:`L` is a *characteristic length* (in our case,
    the conductor diameter) and :math:`\nu` is the kinematic viscosity.

    Parameters
    ----------
    wind_speed:
        :math:`v~\left[\text{m}~\text{s}^{-1}\right]`. The wind speed.
    conductor_diameter:
        :math:`D~\left[\text{m}\right]`. Outer diameter of the conductor.
    kinematic_viscosity_of_air:
        :math:`\nu_f~\left[\text{m}^2~\text{s}^{-1}\right]`. The kinematic viscosity of air.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Re}`. The Reynolds number.
    """
    v = wind_speed
    D = conductor_diameter
    nu_f = kinematic_viscosity_of_air
    return v * D / nu_f


def compute_grashof_number(
    conductor_diameter: Meter,
    surface_temperature: Celsius,
    air_temperature: Celsius,
    kinematic_viscosity_of_air: SquareMeterPerSecond,
    coefficient_of_gravity: MeterPerSquareSecond = 9.807,
) -> Unitless:
    r"""Compute the Grashof number.

    Defined in the nomenclature on page 7 of :cite:p:`cigre601`
    and on page 5 of :cite:p:`cigre207`.

    The Grashof number is a dimensionless quantity that can be used to assess the degree of free
    and forced convective heat transfer.

    Parameters
    ----------
    conductor_diameter:
        :math:`D~\left[\text{m}\right]`. Outer diameter of the conductor.
    surface_temperature:
        :math:`T_s~\left[^\circ\text{C}\right]`. The conductor surface temperature.
    air_temperature:
        :math:`T_a~\left[^\circ\text{C}\right]`. The ambient air temperature.
    kinematic_viscosity_of_air:
        :math:`\nu_f~\left[\text{m}^2~\text{s}^{-1}\right]`. The kinematic viscosity of air.
    coefficient_of_gravity:
        :math:`g~\left[\text{m}~\text{s}^{-2}\right]`. The graviatational constant, optional
        (default=9.807).

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Gr}`. The Grashof number.
    """
    T_s = surface_temperature
    T_a = air_temperature
    T_f = 0.5 * (T_s + T_a)
    D = conductor_diameter
    nu_f = kinematic_viscosity_of_air
    g = coefficient_of_gravity

    return (D**3) * np.abs((T_s - T_a)) * g / ((T_f + 273.15) * (nu_f**2))


def compute_prandtl_number(
    thermal_conductivity_of_air: WattPerMeterPerKelvin,
    dynamic_viscosity_of_air: KilogramPerMeterPerSecond,
    specific_heat_capacity_of_air: JoulePerKilogramPerKelvin,
) -> Unitless:
    r"""Compute the Prandtl number.

    Defined in the nomenclature on page 8 of :cite:p:`cigre601`.

    The Prandtl number measures the ratio between viscosity and thermal diffusivity for a fluid.

    Parameters
    ----------
    thermal_conductivity_of_air:
        :math:`\lambda_f~\left[\text{W}~\text{m}^{-1}~\text{K}^{-1}\right]`. The thermal
        conductivity of air at the given temperature.
    dynamic_viscosity_of_air:
        :math:`\mu_f~\left[\text{kg}~\text{m}^{-1}~\text{s}^{-1}\right]`. The dynamic viscosity of
        air.
    specific_heat_capacity_of_air:
        :math:`\text{J}~\left[\text{kg}^{-1}~\text{K}^{-1}\right]`. The specific heat capacity of
        air.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Pr}`. The Prandtl number.
    """
    lambda_f = thermal_conductivity_of_air
    mu_f = dynamic_viscosity_of_air
    c_f = specific_heat_capacity_of_air

    return c_f * mu_f / lambda_f


def compute_conductor_roughness(
    conductor_diameter: Meter,
    outer_layer_strand_diameter: Meter,
) -> Unitless:
    r"""Compute the surface roughness of the conductor.

    Defined in the text on page 25 of :cite:p:`cigre601` and on page 6 of :cite:p:`cigre207`.

    Parameters
    ----------
    conductor_diameter:
        :math:`D~\left[\text{m}\right]`. Outer diameter of the conductor.
    outer_layer_strand_diameter:
        :math:`d~\left[\text{m}\right]`. The diameter of the strands in the outer layer of the
        conductor.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Rs}`. The roughness number
    """
    D = conductor_diameter
    d = outer_layer_strand_diameter

    if np.any(d < 0):
        raise ValueError(
            dedent(
                """\
            Cannot have negative outer layer strand diameter. If it was set this way to signify
            that it is a smooth conductor, then set the outer layer strand diameter to nan or zero
            instead.\
            """
            )
        )
    if np.any(d >= D):
        raise ValueError(
            "The outer layer strand diameter must be strictly smaller than the conductor diameter."
        )

    return d / (2 * (D - d))



import numpy as np

from ..units import Radian, Unitless


def switch_cos_sin(cos_or_sin: Unitless) -> Unitless:
    return np.sqrt(1 - cos_or_sin**2)


def compute_angle_of_attack(angle_1: Radian, angle_2: Radian) -> Radian:
    angle_diff = np.abs(angle_1 - angle_2) % np.pi
    return np.minimum(angle_diff, np.pi - angle_diff)



from numbers import Real

import numpy as np

from linerate.equations import dimensionless, ieee738, math, solar_angles
from linerate.models.thermal_model import ThermalModel, _copy_method_docstring
from linerate.types import Span, Weather
from linerate.units import Ampere, Celsius, Date, OhmPerMeter, WattPerMeter


class IEEE738(ThermalModel):
    def __init__(
        self,
        span: Span,
        weather: Weather,
        time: Date,
        max_reynolds_number: Real = 50_000,  # Max Reynolds number for forced convection
    ):
        super().__init__(span, weather)
        self.time = time
        self.max_reynolds_number = max_reynolds_number

    @_copy_method_docstring(ThermalModel)
    def compute_resistance(self, conductor_temperature: Celsius, current: Ampere) -> OhmPerMeter:
        return super().compute_resistance(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_joule_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_joule_heating(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_solar_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        alpha_s = self.span.conductor.solar_absorptivity  # alpha in IEEE
        phi = self.span.latitude  # Lat in IEEE
        gamma_c = self.span.conductor_azimuth  # Z_l i IEEE
        y = self.span.conductor_altitude  # H_e in IEEE
        D = self.span.conductor.conductor_diameter  # D_0 in IEEE

        omega = solar_angles.compute_hour_angle_relative_to_noon(self.time, self.span.longitude)
        delta = solar_angles.compute_solar_declination(self.time)
        sin_H_c = solar_angles.compute_sin_solar_altitude(phi, delta, omega)
        Q_s = ieee738.solar_heating.compute_total_heat_flux_density(sin_H_c, True)
        K_solar = ieee738.solar_heating.compute_solar_altitude_correction_factor(y)
        Q_se = ieee738.solar_heating.compute_elevation_correction_factor(K_solar, Q_s)
        chi = solar_angles.compute_solar_azimuth_variable(phi, delta, omega)
        C = solar_angles.compute_solar_azimuth_constant(chi, omega)
        Z_c = solar_angles.compute_solar_azimuth(C, chi)
        cos_theta = solar_angles.compute_cos_solar_effective_incidence_angle(sin_H_c, Z_c, gamma_c)

        return ieee738.solar_heating.compute_solar_heating(alpha_s, Q_se, cos_theta, D)

    @_copy_method_docstring(ThermalModel)
    def compute_convective_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        D = self.span.conductor.conductor_diameter  # D_0 in IEEE
        y = self.span.conductor_altitude  # H_e in IEEE
        V = self.weather.wind_speed  # V_w in IEEE
        T_a = self.weather.air_temperature
        T_c = conductor_temperature
        T_f = 0.5 * (T_c + T_a)  # T_film in IEEE

        mu_f = ieee738.convective_cooling.compute_dynamic_viscosity_of_air(T_f)
        rho_f = ieee738.convective_cooling.compute_air_density(T_f, y)
        nu_f = ieee738.convective_cooling.compute_kinematic_viscosity_of_air(mu_f, rho_f)
        Re = np.minimum(
            dimensionless.compute_reynolds_number(V, D, nu_f),  # N_Re in IEEE
            self.max_reynolds_number,
        )
        delta = math.compute_angle_of_attack(
            self.weather.wind_direction, self.span.conductor_azimuth
        )  # Phi in IEEE
        K_angle = ieee738.convective_cooling.compute_wind_direction_factor(delta)
        k_f = ieee738.convective_cooling.compute_thermal_conductivity_of_air(T_f)
        q_cf = ieee738.convective_cooling.compute_forced_convection(K_angle, Re, k_f, T_c, T_a)
        q_cn = ieee738.convective_cooling.compute_natural_convection(rho_f, D, T_c, T_a)
        return ieee738.convective_cooling.compute_convective_cooling(q_cf, q_cn)

    @_copy_method_docstring(ThermalModel)
    def compute_radiative_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_radiative_cooling(
            conductor_temperature=conductor_temperature, current=current
        )



import warnings

import numpy as np

from ...units import (
    Celsius,
    KilogramPerCubeMeter,
    Meter,
    MeterPerSecond,
    Radian,
    SquareMeterPerSecond,
    Unitless,
    WattPerMeterPerKelvin,
)

# Physical quantities
#####################


def compute_thermal_conductivity_of_air(film_temperature: Celsius) -> WattPerMeterPerKelvin:
    r"""Approximation of the thermal conductivity of air.

    On page 5 of :cite:p:`cigre207`.

    Parameters
    ----------
    film_temperature:
        :math:`T_f = 0.5 (T_s + T_a)~\left[^\circ\text{C}\right]`. The temperature of the
        thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor sufrace temperature.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\lambda_f~\left[\text{W}~\text{m}^{-1}~\text{K}^{-1}\right]`. The thermal
        conductivity of air at the given temperature.
    """
    T_f = film_temperature
    return 2.42e-2 + 7.2e-5 * T_f


def compute_relative_air_density(height_above_sea_level: Meter) -> Unitless:
    r"""Approximation of the relative density of air at a given altitude,
    relative to density at sea level.

    Equation on page 6 of :cite:p:`cigre207`.

    Parameters
    ----------
    height_above_sea_level:
        :math:`y~\left[\text{m}\right]`. The conductor's altitude.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\rho_r`. The relative mass density of air.
    """
    y = height_above_sea_level
    return np.exp(-1.16e-4 * y)


def compute_kinematic_viscosity_of_air(film_temperature: Celsius) -> KilogramPerCubeMeter:
    r"""Approximation of the kinematic viscosity of air at a given temperature.

    Equation on page 5 of :cite:p:`cigre207`.

    Parameters
    ----------
    film_temperature:
        :math:`T_f = 0.5 (T_s + T_a)~\left[^\circ\text{C}\right]`. The temperature of the
        thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor sufrace temperature.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\nu_f~\left[\text{m}^2~\text{s}^{-1}\right]`. The kinematic viscosity of air.
    """
    T_f = film_temperature
    return 1.32e-5 + 9.5e-8 * T_f


def compute_prandtl_number(
    film_temperature: Celsius,
) -> Unitless:
    r"""Compute the Prandtl number.

    Defined on page 5 of :cite:p:`cigre207`.

    The Prandtl number measures the ratio between viscosity and thermal diffusivity for a fluid.

    Parameters
    ----------
    film_temperature:
        :math:`T_f = 0.5 (T_s + T_a)~\left[^\circ\text{C}\right]`. The temperature of the
        thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor sufrace temperature.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Pr}`. The Prandtl number.
    """
    return 0.715 - 2.5e-4 * film_temperature


def compute_reynolds_number(
    wind_speed: MeterPerSecond,
    conductor_diameter: Meter,
    kinematic_viscosity_of_air: SquareMeterPerSecond,
    relative_air_density: Unitless,
) -> Unitless:
    r"""Compute the Reynolds number using the conductor diameter as characteristic length scale.

    Defined on page 5 of :cite:p:`cigre207`.
    This is a non-standard definition which seems to indicate that the kinematic viscosity has to
    be corrected for the density.

    Parameters
    ----------
    wind_speed:
        :math:`v~\left[\text{m}~\text{s}^{-1}\right]`. The wind speed.
    conductor_diameter:
        :math:`D~\left[\text{m}\right]`. Outer diameter of the conductor.
    kinematic_viscosity_of_air:
        :math:`\nu_f~\left[\text{m}^2~\text{s}^{-1}\right]`. The kinematic viscosity of air.
    relative_air_density:
        :math:`\rho_r~1`. The air density relative to density at sea level.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Re}`. The Reynolds number.
    """
    v = wind_speed
    D = conductor_diameter
    nu_f = kinematic_viscosity_of_air
    rho_r = relative_air_density
    return rho_r * v * D / nu_f


## Nusselt number calculation
#############################


def _check_perpendicular_flow_nusseltnumber_out_of_bounds(reynolds_number):
    Re = reynolds_number
    if np.any(Re < 0):
        raise ValueError("Reynolds number cannot be negative.")

    if np.any(np.logical_or(Re < 100, Re > 5e4)):
        warnings.warn("Reynolds number is out of bounds", stacklevel=5)


def compute_perpendicular_flow_nusseltnumber(
    reynolds_number: Unitless,
    conductor_roughness: Meter,
) -> Unitless:
    r"""Compute the Nusselt number for perpendicular flow.

    The Nusselt number is the ratio of conductive heat transfer to convective heat transfer.

    Parameters
    ----------
    reynolds_number:
        :math:`\text{Re}`. The Reynolds number.
    conductor_roughness:
        :math:`\text{Rs}`. The roughness number

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Nu}_{90}`. The perpendicular flow Nusselt number.
    """
    _check_perpendicular_flow_nusseltnumber_out_of_bounds(reynolds_number)
    # From table on page 6 in Cigre207
    Re = reynolds_number
    Rs = conductor_roughness
    conditions = [Re < 100, Re < 2.65e3, Rs <= 0.05]
    B_choices = [0, 0.641, 0.178]
    n_choices = [0, 0.471, 0.633]
    B = np.select(conditions, B_choices, default=0.048)
    n = np.select(conditions, n_choices, default=0.800)
    return B * Re**n


def compute_low_wind_speed_nusseltnumber(
    perpendicular_flow_nusselt_number: Unitless,
) -> Unitless:
    r"""Compute the corrected Nusselt number for low wind speed.

    Parameters
    ----------
    perpendicular_flow_nusselt_number:
        :math:`\text{Nu}_{90}`. The perpendicular flow Nusselt number.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Nu}_{cor}`. The corrected Nusselt number for low wind speed.
    """
    return 0.55 * perpendicular_flow_nusselt_number


def correct_wind_direction_effect_on_nusselt_number(
    perpendicular_flow_nusselt_number: Unitless,
    angle_of_attack: Radian,
) -> Unitless:
    r"""Correct the Nusselt number for the wind's angle-of-attack.

    Equation (14) on page 7 of :cite:p:`cigre207`.

    The perpendicular flow nusselt number is denoted as :math:`\text{Nu}_\delta` in
    :cite:p:`cigre207` since the wind's angle of attack is :math:`\delta`.

    Parameters
    ----------
    perpendicular_flow_nusselt_number:
        :math:`\text{Nu}_{90}`. The perpendicular flow Nusselt number.
    angle_of_attack:
        :math:`\delta~\left[\text{radian}\right]`. The wind angle-of-attack.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Nu}_\delta`. The Nusselt number for the given wind angle-of-attack.
    """
    delta = angle_of_attack
    Nu_90 = perpendicular_flow_nusselt_number
    sin_delta = np.sin(delta)
    # Equation (14) on page 7 of Cigre207
    return (
        np.where(
            delta <= np.radians(24),
            0.42 + 0.68 * (sin_delta**1.08),
            0.42 + 0.58 * (sin_delta**0.90),
        )
        * Nu_90
    )


## Natural convection computations (no wind):
#############################################


def _check_horizontal_natural_nusselt_number(
    grashof_number: Unitless, prandtl_number: Unitless
) -> None:
    GrPr = grashof_number * prandtl_number
    if np.any(GrPr < 0):
        raise ValueError("GrPr cannot be negative.")
    elif np.any(GrPr >= 1e12):
        raise ValueError("GrPr out of bounds: Must be < 10^12.")


def compute_horizontal_natural_nusselt_number(
    grashof_number: Unitless,
    prandtl_number: Unitless,
) -> Unitless:
    r"""The Nusselt number for natural (passive) convection on a horizontal conductor.

    Equation (16) and Table II on page 7 of :cite:p:`cigre207`.
    We expand the allowable range by using Table 9.1 of :cite:p:`incropera2007`.

    Parameters
    ----------
    grashof_number:
        :math:`\text{Gr}`. The Grashof number.
    prandtl_number:
        :math:`\text{Pr}`. The Prandtl number.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{Nu}_0`. The natural convection nusselt number assuming horizontal conductor.
    """
    _check_horizontal_natural_nusselt_number(grashof_number, prandtl_number)
    GrPr = grashof_number * prandtl_number
    # CIGRE 207 only allows GrPr in the range 1e2-1e6.
    # In Incropera (2006), Table 9.1, the same values appear, but the table covers a wider range
    # from 1e-10 to 1e12, which we use here
    conditions = [GrPr < 1e-2, GrPr < 1e2, GrPr < 1e4, GrPr < 1e7, GrPr < 1e12]
    n_choices = [0.058, 0.148, 0.188, 0.250, 0.333]
    C_choices = [0.0675, 1.02, 0.850, 0.480, 0.125]
    C = np.select(conditions, C_choices, default=np.nan)
    n = np.select(conditions, n_choices, default=np.nan)
    return C * GrPr**n


def compute_nusselt_number(
    forced_convection_nusselt_number: Unitless,
    natural_nusselt_number: Unitless,
    low_wind_nusselt_number: Unitless,
    wind_speed: MeterPerSecond,
) -> Unitless:
    r"""Compute the nusselt number.

    Described on page 7 of :cite:p:`cigre207`.

    Parameters
    ----------
    forced_convection_nusselt_number:
        :math:`\text{Nu}_\delta`. The Nusselt number for the given wind angle-of-attack.
    natural_nusselt_number:
        :math:`\text{Nu}_\delta`. The natural convection nusselt number for horizontal conductor.
    low_wind_nusselt_number:
        :math:`\text{Nu}_cor`. Corrected Nusselt number for low wind.
    wind_speed:
        :math:`v~\left[\text{m}~\text{s}^{-1}\right]`. The wind speed.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`Nu`. The nusselt number.
    """
    normal_nusselt = np.maximum(forced_convection_nusselt_number, natural_nusselt_number)
    low_wind_nusselt = np.maximum(normal_nusselt, low_wind_nusselt_number)
    low_wind_speed = wind_speed < 0.5
    return np.where(low_wind_speed, low_wind_nusselt, normal_nusselt)



from functools import partial
from typing import Callable

import numpy as np

from .units import Ampere, Celsius, FloatOrFloatArray, WattPerMeter

__all__ = ["bisect", "compute_conductor_temperature", "compute_conductor_ampacity"]


def bisect(
    f: Callable[[FloatOrFloatArray], FloatOrFloatArray],
    xmin: FloatOrFloatArray,
    xmax: FloatOrFloatArray,
    tolerance: float,
) -> FloatOrFloatArray:
    r"""Compute the roots of a function using a vectorized bisection method.

    Parameters
    ----------
    f:
        :math:`f: \mathbb{R}^n \to \mathbb{R}^n`. Function whose roots we wish to find.
    xmin:
        :math:`x_\min`. Minimum value for the free parameter, :math:`\mathbf{x}`. It is required
        that :math:`\text{sign}(f_i(x_\min)) \neq \text{sign}(f_i(x_\max))`. for all :math:`i`.
    xmax:
        :math:`x_\max`. Maximum value for the free parameter, :math:`\mathbf{x}`. It is required
        that :math:`\text{sign}(f_i(x_\min)) \neq \text{sign}(f_i(x_\max))`. for all :math:`i`.
    tolerance:
        :math:`\Delta x`. The bisection iterations will terminate once all :math:`x_i`-s are
        bounded within an interval of size :math:`\Delta x` or less. The bisection method will
        run for :math:`\left\lceil\frac{x_\max - x_\min}{\Delta x}\right\rceil`
        iterations.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\tilde{\mathbf{x}}`. Estimate for the roots of :math:`f`. For each :math:`f_i`,
        there is a root :math:`x_i \in [\tilde{x}_i - 0.5 \Delta x, \tilde{x}_i + 0.5 \Delta x]`
        so :math:`f_i(x_i) = 0`.
    """
    if not np.all(np.isfinite(xmin)) or not np.all(np.isfinite(xmax)):
        raise ValueError("xmin and xmax must be finite.")
    interval = np.max(np.abs(xmax - xmin))

    f_left = f(xmin)
    f_right = f(xmax)

    invalid_mask = np.sign(f_left) == np.sign(f_right)
    if np.any(invalid_mask):
        raise ValueError(
            "f(xmin) and f(xmax) have the same sign. Consider increasing the search interval."
        )

    while interval > tolerance:
        xmid = 0.5 * (xmax + xmin)
        interval *= 0.5
        f_mid = f(xmid)

        mask = f_mid * f_left > 0  # fast way to check sign(f_mid) == sign(f_left)
        xmin = np.where(mask, xmid, xmin)
        xmax = np.where(mask, xmax, xmid)
        f_left = np.where(mask, f_mid, f_left)
        f_right = np.where(mask, f_right, f_mid)

    out = 0.5 * (xmax + xmin)
    return out


def compute_conductor_temperature(
    heat_balance: Callable[[Celsius, Ampere], WattPerMeter],
    current: Ampere,
    min_temperature: Celsius = -30,
    max_temperature: Celsius = 150,
    tolerance: float = 0.5,  # Celsius
) -> Celsius:
    r"""Use the bisection method to compute the steady state conductor temperature.

    Parameters
    ----------
    heat_balance:
        :math:`f(T, A) = P_J + P_s - P_c - P_r~\left[\text{W}~\text{m}^{-1}\right]`. A function of
        both temperature and current that returns the heat balance for the given
        temperature-current pair.
    current:
        :math:`I_\text{max}~\left[\text{A}\right]`. The current flowing through the conductor.
    min_temperature:
        :math:`T_\text{min}~\left[^\circ\text{C}\right]`. Lower bound for the numerical scheme for
        computing the temperature
    max_temperature:
        :math:`T_\text{max}~\left[^\circ\text{C}\right]`. Upper bound for the numerical scheme for
        computing the temperature
    tolerance:
        :math:`\Delta T~\left[^\circ\text{C}\right]`. The numerical accuracy of the
        temperature. The bisection iterations will stop once the numerical temperature
        uncertainty is below :math:`\Delta T`. The bisection method will run for
        :math:`\left\lceil\frac{T_\text{min} - T_\text{min}}{\Delta T}\right\rceil` iterations.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`I~\left[\text{A}\right]`. The thermal rating.
    """
    f = partial(heat_balance, current=current)

    return bisect(f, min_temperature, max_temperature, tolerance)


def compute_conductor_ampacity(
    heat_balance: Callable[[Celsius, Ampere], WattPerMeter],
    max_conductor_temperature: Celsius,
    min_ampacity: Ampere = 0,
    max_ampacity: Ampere = 5_000,
    tolerance: float = 1,  # Ampere
) -> Ampere:
    r"""Use the bisection method to compute the steady-state thermal rating (ampacity).

    Parameters
    ----------
    heat_balance:
        :math:`f(T, A) = P_J + P_s - P_c - P_r~\left[\text{W}~\text{m}^{-1}\right]`. A function of
        both temperature and current that returns the heat balance for the given
        temperature-current pair.
    max_conductor_temperature:
        :math:`T_\text{max}~\left[^\circ\text{C}\right]`. Maximum allowed conductor temperature
    min_ampacity:
        :math:`I_\text{min}~\left[\text{A}\right]`. Lower bound for the numerical scheme for
        computing the ampacity
    max_ampacity:
        :math:`I_\text{min}~\left[\text{A}\right]`. Upper bound for the numerical scheme for
        computing the ampacity
    tolerance:
        :math:`\Delta I~\left[\text{A}\right]`. The numerical accuracy of the ampacity. The
        bisection iterations will stop once the numerical ampacity uncertainty is below
        :math:`\Delta I`. The bisection method will run for
        :math:`\left\lceil\frac{I_\text{max} - I_\text{min}}{\Delta I}\right\rceil` iterations.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`I~\left[\text{A}\right]`. The thermal rating.
    """
    f = partial(heat_balance, max_conductor_temperature)

    return bisect(f, min_ampacity, max_ampacity, tolerance)



import numpy as np

from linerate.units import Celsius, Unitless, WattPerMeter, WattPerMeterPerKelvin


def compute_convective_cooling(
    surface_temperature: Celsius,
    air_temperature: Celsius,
    nusselt_number: Unitless,
    thermal_conductivity_of_air: WattPerMeterPerKelvin,
) -> WattPerMeter:
    r"""Compute the convective cooling of the conductor.

    Equation (17) on page 24 of :cite:p:`cigre601`
    and Equation (12) on page 6 of :cite:p:`cigre207`.

    Parameters
    ----------
    surface_temperature:
        :math:`T_s~\left[^\circ\text{C}\right]`. The conductor surface temperature.
    air_temperature:
        :math:`T_a~\left[^\circ\text{C}\right]`. The ambient air temperature.
    nusselt_number:
        :math:`Nu`. The nusselt number.
    thermal_conductivity_of_air:
        :math:`\lambda_f~\left[\text{W}~\text{m}^{-1}~\text{K}^{-1}\right]`. The thermal
        conductivity of air at the given temperature.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`P_c~\left[\text{W}~\text{m}^{-1}\right]`. The convective cooling of the conductor.
        Either due to wind, or passive convection, whichever is largest.
    """
    pi = np.pi
    lambda_f = thermal_conductivity_of_air
    T_s = surface_temperature
    T_a = air_temperature
    Nu = nusselt_number

    return pi * lambda_f * (T_s - T_a) * Nu



from typing import Union

import numpy as np

from ..units import (
    Ampere,
    Celsius,
    OhmPerMeter,
    SquareMeter,
    SquareMeterPerAmpere,
    Unitless,
    WattPerMeter,
)


def compute_resistance(
    conductor_temperature: Celsius,
    temperature1: Celsius,
    temperature2: Celsius,
    resistance_at_temperature1: OhmPerMeter,
    resistance_at_temperature2: OhmPerMeter,
) -> OhmPerMeter:
    r"""Compute the (possibly AC-)resistance of the conductor at a given temperature.

    The resistance is linearly interpolated/extrapolated based on the two temperature-resistance
    measurement pairs provided as arguments.

    Parameters
    ----------
    conductor_temperature:
        :math:`T~\left[^\circ\text{C}\right]`. The average conductor temperature.
    temperature1:
        :math:`T_1~\left[^\circ\text{C}\right]`. The first temperature measurement.
    temperature2:
        :math:`T_2~\left[^\circ\text{C}\right]`. The second temperature measurement.
    resistance_at_temperature1:
        :math:`R_1~\left[\Omega~\text{m}^{-1}\right]`. The resistance at temperature :math:`T=T_1`.
    resistance_at_temperature2:
        :math:`R_2~\left[\Omega~\text{m}^{-1}\right]`. The resistance at temperature :math:`T=T_2`.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`R~\left[\Omega\right]`. The resistance at the given temperature.
    """
    T_av = conductor_temperature
    T_1, T_2 = temperature1, temperature2
    R_1, R_2 = resistance_at_temperature1, resistance_at_temperature2

    a = (R_2 - R_1) / (T_2 - T_1)
    b = R_1 - a * T_1
    return a * T_av + b


def correct_resistance_acsr_magnetic_core_loss(
    ac_resistance: OhmPerMeter,
    current: Ampere,
    aluminium_cross_section_area: SquareMeter,
    constant_magnetic_effect: Union[Unitless, None],
    current_density_proportional_magnetic_effect: Union[SquareMeterPerAmpere, None],
    max_relative_increase: Unitless,
) -> OhmPerMeter:
    r"""Correct for extra resistance in ACSR conductors due to magnetic effects in the steel core.

    Aluminium-conductor steel-reinforced (ACSR) conductors have an additional resistance due to
    magnetic effects in the steel core. Particularly conductors with an odd number of layers with
    aluminium wires.

    According to :cite:p:`ieee.acsr.taskforce`, we can assume a linear relationship between the
    current density and the relative increase in resistance due to the steel core of three-layer
    ACSR. However, the task force also says that we can assume that the increase saturates at
    6%. In :cite:p:`cigre345`, it is stated that the maximum increase for three-layer ACSR is 5%.

    For ACSR with an even number of layers, the effect is negligible since we get cancelling
    magnetic fields, and for mono-layer ACSR, the effect behaves differently. Still, some software
    providers use the same correction scheme for mono-layer ACSR, but with a higher saturation
    point (typically 20%, since that is the maximum resistance increase in mono-layer ACSR
    :cite:p:`cigre345`).

    The linear but saturating increase in resistance leads to the following correction scheme

    .. math::

        R_\text{corrected} = R \min(c_\text{max}, b + mJ),

    where :math:`R_\text{corrected}` is the corrected AC resistance for ACSR conductors,
    :math:`R` is the uncorrected value for the AC resistance at the given temperature,
    :math:`c_\text{max}` is the maximum relative increase (e.g. 6% or :math:`1.06`` for
    three-layer ACSR conductors), :math:`J` is the current density (current divided by the
    aluminium cross-section area, :math:`I/A`) and :math:`b` and :math:`m` are the constant and
    current-density proprtional magnetic effects, respectively (for example obtained from linear
    regression curve).

    Parameters
    ----------
    ac_resistance:
        :math:`R~\left[\Omega\right]`. The AC resistance of the conductor.
    current:
        :math:`I~\left[\text{A}\right]`. The current going through the conductor.
    aluminium_cross_section_area:
        :math:`A_{\text{Al}}~\left[\text{m}^2\right]`. The cross sectional area of the aluminium
        strands in the conductor.
    constant_magnetic_effect:
        :math:`b`. The constant magnetic effect, most likely equal to 1. If ``None``, then no
        correction is used (useful for non-ACSR cables).
    current_density_proportional_magnetic_effect:
        :math:`m`. The current density proportional magnetic effect. If ``None``, then it is
        assumed equal to 0.
    max_relative_increase:
        :math:`c_\text{max}`. Saturation point of the relative increase in conductor resistance.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`R_\text{corrected}~\left[\Omega\right]`. The resistance of the conductor after
        taking steel core magnetization effects into account.
    """
    I = current  # noqa
    R = ac_resistance
    A = aluminium_cross_section_area
    b = constant_magnetic_effect
    m = current_density_proportional_magnetic_effect

    if b is None:
        return R

    if m is None or np.all(m == 0):
        return b * R

    J = I / A
    return np.minimum(b + J * m, max_relative_increase) * R


def compute_joule_heating(current: Ampere, resistance: OhmPerMeter) -> WattPerMeter:
    r"""Compute the Joule heating, assuming AC-resistance for AC lines.

    Parameters
    ----------
    current:
        :math:`I~\left[\text{A}\right]`. The current going through the conductor.
    Resistance:
        :math:`R~\left[\Omega\right]`. The (possibly AC-)resistance of the conductor, correcting
        for all possible magnetisation effects.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`P_J~\left[\text{W}~\text{m}^{-1}\right]`. The Joule heating of the conductor
    """
    I = current  # noqa
    R = resistance
    return I * I * R



from abc import ABC, abstractmethod
from typing import Dict

from linerate import solver
from linerate.equations import joule_heating, radiative_cooling
from linerate.types import Span, Weather
from linerate.units import Ampere, Celsius, OhmPerMeter, WattPerMeter


def _copy_method_docstring(parent_class):
    def inner(func):
        func.__doc__ = getattr(parent_class, func.__name__).__doc__
        return func

    return inner


class ThermalModel(ABC):
    """Abstract class for a minimal conductor thermal model."""

    @abstractmethod
    def __init__(self, span: Span, weather: Weather):
        self.span = span
        self.weather = weather

    @abstractmethod
    def compute_resistance(self, conductor_temperature: Celsius, current: Ampere) -> OhmPerMeter:
        r"""Compute the conductor resistance, :math:`R~\left[\Omega~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`R~\left[\Omega\right]`. The resistance at the given temperature and current.
        """
        resistance = joule_heating.compute_resistance(
            conductor_temperature,
            temperature1=self.span.conductor.temperature1,
            temperature2=self.span.conductor.temperature2,
            resistance_at_temperature1=self.span.conductor.resistance_at_temperature1,
            resistance_at_temperature2=self.span.conductor.resistance_at_temperature2,
        )

        A = self.span.conductor.aluminium_cross_section_area
        b = self.span.conductor.constant_magnetic_effect
        m = self.span.conductor.current_density_proportional_magnetic_effect
        max_increase = self.span.conductor.max_magnetic_core_relative_resistance_increase

        return joule_heating.correct_resistance_acsr_magnetic_core_loss(
            ac_resistance=resistance,
            current=current,
            aluminium_cross_section_area=A,
            constant_magnetic_effect=b,
            current_density_proportional_magnetic_effect=m,
            max_relative_increase=max_increase,
        )

    @abstractmethod
    def compute_joule_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        r"""Compute the Joule heating, :math:`P_J~\left[\text{W}~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_J~\left[\text{W}~\text{m}^{-1}\right]`. The Joule heating.
        """
        resistance = self.compute_resistance(conductor_temperature, current)
        return joule_heating.compute_joule_heating(current, resistance)

    @abstractmethod
    def compute_solar_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        r"""Compute the solar heating, :math:`P_S~\left[\text{W}~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_s~\left[\text{W}~\text{m}^{-1}\right]`. The solar heating.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_convective_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        r"""Compute the convective cooling, :math:`P_c~\left[\text{W}~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_c~\left[\text{W}~\text{m}^{-1}\right]`. The convective cooling.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_radiative_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        r"""Compute the radiative cooling, :math:`P_r~\left[\text{W}~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_r~\left[\text{W}~\text{m}^{-1}\right]`. The radiative cooling heating.
        """
        return radiative_cooling.compute_radiative_cooling(
            surface_temperature=conductor_temperature,
            air_temperature=self.weather.air_temperature,
            conductor_diameter=self.span.conductor.conductor_diameter,
            conductor_emissivity=self.span.conductor.emissivity,
        )

    def compute_heat_balance(self, conductor_temperature: Celsius, current: Ampere) -> WattPerMeter:
        r"""Compute the conductor's heat balance. Positive means heating and negative means cooling.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_J + P_s - P_c - P_r~\left[\text{W}~\text{m}^{-1}\right]`. The heat balance.
        """
        P_j = self.compute_joule_heating(conductor_temperature, current)
        P_s = self.compute_solar_heating(conductor_temperature, current)
        P_c = self.compute_convective_cooling(conductor_temperature, current)
        P_r = self.compute_radiative_cooling(conductor_temperature, current)
        return P_j + P_s - P_c - P_r

    def compute_info(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> Dict[str, WattPerMeter]:
        r"""Create a dictionary with the different heating and cooling effects.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Dict[str, WattPerMeter]
            A dictionary with the magnitude of the different heating and cooling effects.
        """
        return {
            "convective_cooling": self.compute_convective_cooling(conductor_temperature, current),
            "radiative_cooling": self.compute_radiative_cooling(conductor_temperature, current),
            "joule_heating": self.compute_joule_heating(conductor_temperature, current),
            "solar_heating": self.compute_solar_heating(conductor_temperature, current),
        }

    def compute_steady_state_ampacity(
        self,
        max_conductor_temperature: Celsius,
        min_ampacity: Ampere = 0,
        max_ampacity: Ampere = 5000,
        tolerance: float = 1.0,
    ) -> Ampere:
        r"""Use the bisection method to compute the steady-state thermal rating (ampacity).

        Parameters
        ----------
        max_conductor_temperature:
            :math:`T_\text{max}~\left[^\circ\text{C}\right]`. Maximum allowed conductor temperature
        min_ampacity:
            :math:`I_\text{min}~\left[\text{A}\right]`. Lower bound for the numerical scheme for
            computing the ampacity
        max_ampacity:
            :math:`I_\text{min}~\left[\text{A}\right]`. Upper bound for the numerical scheme for
            computing the ampacity
        tolerance:
            :math:`\Delta I~\left[\text{A}\right]`. The numerical accuracy of the ampacity. The
            bisection iterations will stop once the numerical ampacity uncertainty is below
            :math:`\Delta I`. The bisection method will run for
            :math:`\left\lceil\frac{I_\text{min} - I_\text{min}}{\Delta I}\right\rceil` iterations.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`I~\left[\text{A}\right]`. The thermal rating.
        """
        I = solver.compute_conductor_ampacity(  # noqa
            self.compute_heat_balance,
            max_conductor_temperature=max_conductor_temperature,
            min_ampacity=min_ampacity,
            max_ampacity=max_ampacity,
            tolerance=tolerance,
        )
        n = self.span.num_conductors
        return I * n

    def compute_conductor_temperature(
        self,
        current: Ampere,
        min_temperature: Celsius = -30,
        max_temperature: Celsius = 150,
        tolerance: float = 0.5,
    ) -> Celsius:
        r"""Use the bisection method to compute the steady state conductor temperature.

        Parameters
        ----------
        current:
            :math:`I_\text{max}~\left[\text{A}\right]`. The current flowing through the conductor.
        min_temperature:
            :math:`T_\text{min}~\left[^\circ\text{C}\right]`. Lower bound for the numerical scheme
            for computing the temperature
        max_temperature:
            :math:`T_\text{max}~\left[^\circ\text{C}\right]`. Upper bound for the numerical scheme
            for computing the temperature
        tolerance:
            :math:`\Delta T~\left[^\circ\text{C}\right]`. The numerical accuracy of the
            temperature. The bisection iterations will stop once the numerical temperature
            uncertainty is below :math:`\Delta T`. The bisection method will run for
            :math:`\left\lceil\frac{T_\text{min} - T_\text{min}}{\Delta T}\right\rceil` iterations.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`I~\left[\text{A}\right]`. The thermal rating.
        """
        T = solver.compute_conductor_temperature(
            self.compute_heat_balance,
            current=current,
            min_temperature=min_temperature,
            max_temperature=max_temperature,
            tolerance=tolerance,
        )
        n = self.span.num_conductors
        return T / n



from linerate.units import OhmPerMeter


def correct_resistance_for_skin_effect(
    dc_resistance: OhmPerMeter,
) -> OhmPerMeter:
    r"""
    Return resistance with constant correction for skin effect, using simple method from
    Cigre 207, see section 2.1.1.

    Parameters
    ----------
    dc_resistance:
        :math:`R~\left[\Omega\right]`. The DC resistance of the conductor.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`R_\text{corrected}~\left[\Omega\right]`. The resistance of the conductor after
        taking skin effect into account.
    """
    return 1.0123 * dc_resistance


# TODO: Implement section 2.1.2?



import numpy as np
from numpy import pi

from ...units import Meter, Unitless, WattPerSquareMeter


def compute_direct_solar_radiation(
    sin_solar_altitude: Unitless,
    clearness_ratio: Unitless,
    height_above_sea_level: Meter,
) -> WattPerSquareMeter:
    r"""Compute the direct solar radiation.

    Equation (10-11) on page 19 of :cite:p:`cigre601`. Equation (10) states that the direct solar
    radiation on a surface normal to the solar beam at sea level, :math:`I_{B(0)}`, is given by

    .. math::

        N_s \frac{1280 \sin(H_s)}{\sin(H_s) + 0.314},

    where :math:`N_s` is the clearness ratio which is used to adjust the amount of radiation
    compared to what goes through a standard Indian atmosphere, and :math:`H_s` is the solar
    altitude.

    While the solar radiation model is based on :cite:p:`sharma1965interrelationships` and
    therefore have parameters estimated for an Indian atmosphere, it gives comparable results to
    the solar radiation model in the IEEE standard :cite:p:`ieee738`. It is therefore reasonable to
    assume that the parameters work in other climates as well.

    Parameters
    ----------
    sin_solar_altitude:
        :math:`\sin\left(H_s\right)`. The sine of the solar altitude.
    clearness_ratio:
        :math:`N_s`. The clearness ratio (or clearness number in
        :cite:p:`sharma1965interrelationships,cigre207`).
    height_above_sea_level:
        :math:`y~\left[\text{m}\right]`. The conductor's altitude.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`I_B~\left[\text{W}~\text{m}^{-2}\right]`. The direct solar radiation.

    Note
    ----
    The 1280 originates and 0.314 in the above equation originates from
    :cite:p:`sharma1965interrelationships`, which is cited in :cite:p:`morgan1982thermal` (which is
    listed as the reference in :cite:p:`cigre601`). In :cite:p:`sharma1965interrelationships` the
    empirical relationship

    .. math::

        I_{B(0)} = \frac{1.842 \sin(H_s)}{\sin(H_s) + 0.3135}~\text{Ly}~\text{min}^{-1}

    is introduced, and by converting from Langley per minute to :math:`\text{W}~\text{m}^{-2}`, we
    obtain

    .. math::

        I_{B(0)} = N_s \frac{1284.488 \sin(H_s)}{\sin(H_s) + 0.3135}~\text{W}~\text{m}^{-2},

    which is equal to the equation we use (with three significant digits).
    """
    sin_H_s = sin_solar_altitude
    N_s = clearness_ratio
    y = height_above_sea_level

    I_B_0 = N_s * 1280 * sin_H_s / (sin_H_s + 0.314)
    # Equation 19 says that
    # I_B = I_B_0 * (1 + 1.4e-4 * y * (1367/I_B_0 - 1)) # noqa: E800
    # However, if I_B_0 = 0, this will divide by 0. To return NaN-values if and only
    # if the input is NaN, we therefore reformulate it
    scaled_y = 1.4e-4 * y
    I_B = I_B_0 * (1 - scaled_y) + 1367 * scaled_y

    return np.where(sin_H_s >= 0, I_B, 0 * I_B)


def compute_diffuse_sky_radiation(
    direct_solar_radiation: WattPerSquareMeter,
    sin_solar_altitude: Unitless,
) -> WattPerSquareMeter:
    r"""Compute the diffuse radiation (light scattered in the atmosphere).

    Equation (13) on page 20 of :cite:p:`cigre601`.

    This equation differs from :cite:p:`cigre207`, however the difference is small, and the
    diffuse radiation is a small contributor to the overall solar radiation, so the total
    discrepancy between the models is small.

    Parameters
    ----------
    direct_solar_radiation:
        :math:`I_B~\left[\text{W}~\text{m}^{-2}\right]`. The direct solar radiation.
    sin_solar_altitude:
        :math:`\sin\left(H_s\right)`. The sine of the solar altitude.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`I_d~\left[\text{W}~\text{m}^{-2}\right]`.The diffuse solar radiation.
    """
    sin_H_s = sin_solar_altitude
    I_B = direct_solar_radiation
    return np.maximum(0, (430.5 - 0.3288 * I_B)) * np.maximum(0, sin_H_s)


def compute_global_radiation_intensity(
    direct_solar_radiation: WattPerSquareMeter,
    diffuse_sky_radiation: WattPerSquareMeter,
    albedo: Unitless,
    sin_angle_of_sun_on_line: Unitless,
    sin_solar_altitude: Unitless,
) -> WattPerSquareMeter:
    r"""Compute the global radiation intensity experienced by the conductor.

    Equation (9) on page 18 of :cite:p:`cigre601` state that the global radiation intensity,
    :math:`I_T`, is given by

    .. math::

        I_T =
            I_B \left(\sin(\eta) + 0.5 F \pi \sin(H_s)\right) +
            I_d \left(1 + 0.5 F \pi\right),

    where :math:`\eta` is the incidence angle of the sun on the line, :math:`H_s` is the solar
    altitude and :math:`F` is the ground albedo (amount of radiation diffusely reflected from the
    ground). The factor :math:`0.5 \pi` is due the assumption that the ground reflects light
    diffusely and uniformly in all directions, so the reflected energy is always directed normally
    to the line. In CIGRE207, it is also assumed that the diffuse radiation is uniformly directed,
    which leads to :math:`I_d (0.5 \pi + 0.5 F \pi)` instead of  :math:`I_d (1 + 0.5 F \pi)`

    Parameters
    ----------
    direct_solar_radiation:
        :math:`I_B~\left[\text{W}~\text{m}^{-2}\right]`. The direct solar radiation.
    diffuse_sky_radiation:
        :math:`I_d~\left[\text{W}~\text{m}^{-2}\right]`.The diffuse solar radiation.
    albedo:
        :math:`F`. The ground albedo.
    sin_angle_of_sun_on_line:
        :math:`\sin\left(\eta\right)`. The sine of the angle of the sun on the line.
    sin_solar_altitude:
        :math:`\sin\left(H_s\right)`. The sine of the solar altitude.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`I_T~\left[\text{W}~\text{m}^{-2}\right]`. The global radiation intensity.

    Note
    ----
    The following values are given for the albedo in :cite:p:`cigre601`:

    .. list-table::
        :widths: 50 50
        :header-rows: 1

        * - Ground
          - Albedo
        * - Water (:math:`H_s > 30^\circ`)
          - 0.05
        * - Forest
          - 0.1
        * - Urban areas
          - 0.15
        * - Soil, grass and crops
          - 0.2
        * - Sand
          - 0.3
        * - Ice
          - 0.4-0.6
        * - Snow
          - 0.6-0.8
    """
    I_B = direct_solar_radiation
    I_d = diffuse_sky_radiation
    F = albedo
    sin_H_s = sin_solar_altitude
    sin_eta = sin_angle_of_sun_on_line
    F_pi_half = 0.5 * pi * F

    return I_B * (sin_eta + F_pi_half * sin_H_s) + I_d * (1 + F_pi_half)  # type: ignore



from numbers import Real

import numpy as np

from linerate.equations import (
    cigre601,
    convective_cooling,
    dimensionless,
    math,
    solar_angles,
    solar_heating,
)
from linerate.equations.math import switch_cos_sin
from linerate.models.thermal_model import ThermalModel, _copy_method_docstring
from linerate.types import Span, Weather
from linerate.units import (
    Ampere,
    Celsius,
    Date,
    JoulePerKilogramPerKelvin,
    OhmPerMeter,
    WattPerMeter,
)


class Cigre601(ThermalModel):
    def __init__(
        self,
        span: Span,
        weather: Weather,
        time: Date,
        max_reynolds_number: Real = 4000,  # Max value of the angle correction in CIGRE601
    ):
        super().__init__(span, weather)
        self.time = time
        self.max_reynolds_number = max_reynolds_number

    @_copy_method_docstring(ThermalModel)
    def compute_resistance(self, conductor_temperature: Celsius, current: Ampere) -> OhmPerMeter:
        return super().compute_resistance(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_joule_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_joule_heating(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_solar_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        alpha_s = self.span.conductor.solar_absorptivity
        F = self.span.ground_albedo
        phi = self.span.latitude
        gamma_c = self.span.conductor_azimuth
        y = self.span.conductor_altitude
        N_s = self.weather.clearness_ratio
        D = self.span.conductor.conductor_diameter

        omega = solar_angles.compute_hour_angle_relative_to_noon(self.time, self.span.longitude)
        delta = solar_angles.compute_solar_declination(self.time)
        sin_H_s = solar_angles.compute_sin_solar_altitude(phi, delta, omega)
        chi = solar_angles.compute_solar_azimuth_variable(phi, delta, omega)
        C = solar_angles.compute_solar_azimuth_constant(chi, omega)
        gamma_s = solar_angles.compute_solar_azimuth(C, chi)  # Z_c in IEEE
        cos_eta = solar_angles.compute_cos_solar_effective_incidence_angle(
            sin_H_s, gamma_s, gamma_c
        )
        sin_eta = switch_cos_sin(cos_eta)

        I_B = cigre601.solar_heating.compute_direct_solar_radiation(sin_H_s, N_s, y)
        I_d = cigre601.solar_heating.compute_diffuse_sky_radiation(I_B, sin_H_s)
        I_T = cigre601.solar_heating.compute_global_radiation_intensity(
            I_B, I_d, F, sin_eta, sin_H_s
        )
        return solar_heating.compute_solar_heating(
            alpha_s,
            I_T,
            D,
        )

    @_copy_method_docstring(ThermalModel)
    def compute_convective_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        D = self.span.conductor.conductor_diameter
        d = self.span.conductor.outer_layer_strand_diameter
        y = self.span.conductor_altitude
        beta = self.span.inclination
        V = self.weather.wind_speed
        T_a = self.weather.air_temperature
        T_c = conductor_temperature
        T_f = 0.5 * (T_c + T_a)

        # Compute physical quantities
        lambda_f = cigre601.convective_cooling.compute_thermal_conductivity_of_air(T_f)
        mu_f = cigre601.convective_cooling.compute_dynamic_viscosity_of_air(T_f)
        gamma_f = cigre601.convective_cooling.compute_air_density(T_f, y)
        nu_f = cigre601.convective_cooling.compute_kinematic_viscosity_of_air(mu_f, gamma_f)
        c_f: JoulePerKilogramPerKelvin = 1005
        delta = math.compute_angle_of_attack(
            self.weather.wind_direction, self.span.conductor_azimuth
        )

        # Compute unitless quantities
        Re = np.minimum(
            dimensionless.compute_reynolds_number(V, D, nu_f),
            self.max_reynolds_number,
        )
        Gr = dimensionless.compute_grashof_number(D, T_c, T_a, nu_f)
        Pr = dimensionless.compute_prandtl_number(lambda_f, mu_f, c_f)
        Rs = dimensionless.compute_conductor_roughness(D, d)

        # Compute nusselt numbers
        Nu_90 = cigre601.convective_cooling.compute_perpendicular_flow_nusseltnumber(
            reynolds_number=Re, conductor_roughness=Rs
        )
        Nu_delta = cigre601.convective_cooling.correct_wind_direction_effect_on_nusselt_number(
            Nu_90, delta, Rs
        )

        Nu_0 = cigre601.convective_cooling.compute_horizontal_natural_nusselt_number(Gr, Pr)
        Nu_beta = cigre601.convective_cooling.correct_natural_nusselt_number_inclination(
            Nu_0, beta, Rs
        )

        Nu = cigre601.convective_cooling.compute_nusselt_number(
            forced_convection_nusselt_number=Nu_delta, natural_nusselt_number=Nu_beta
        )

        return convective_cooling.compute_convective_cooling(
            surface_temperature=conductor_temperature,
            air_temperature=self.weather.air_temperature,
            nusselt_number=Nu,
            thermal_conductivity_of_air=lambda_f,
        )

    @_copy_method_docstring(ThermalModel)
    def compute_radiative_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_radiative_cooling(
            conductor_temperature=conductor_temperature, current=current
        )

    def compute_temperature_gradient(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> Celsius:
        r"""Estimate the difference between the core temperature and the surface temperature.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`T_c - T_s~\left[^\circ \text{C}\right]`. The difference between the core and the
            surface temperature of the conductor.
        """
        n = self.span.num_conductors
        T_c = conductor_temperature
        I = current / n  # noqa
        R = self.compute_resistance(conductor_temperature=T_c, current=I)
        return cigre601.convective_cooling.compute_temperature_gradient(
            total_heat_gain=I * R,
            conductor_thermal_conductivity=self.span.conductor.thermal_conductivity,  # type: ignore  # noqa
            core_diameter=self.span.conductor.core_diameter,
            conductor_diameter=self.span.conductor.conductor_diameter,
        )



"""
These thermal model classes provide an easy-to-use interface to compute the thermal rating.
They store conductor and span metadata as well as the weather parameters, compute all the
heating and cooling effects and use those to estimate the thermal rating and conductor
temperature. All numerical heavy-lifting is handled by the ``linerate.equations`` and the
``linerate.solver`` modules.
"""

from linerate.models.cigre207 import Cigre207
from linerate.models.cigre601 import Cigre601
from linerate.models.ieee738 import IEEE738
from linerate.models.thermal_model import ThermalModel

__all__ = ["ThermalModel", "Cigre601", "IEEE738", "Cigre207"]



import numpy as np
from numpy import pi

from ...units import Meter, Unitless, WattPerSquareMeter
from .. import cigre601


def compute_direct_solar_radiation(
    sin_solar_altitude: Unitless,
    height_above_sea_level: Meter,
) -> WattPerSquareMeter:
    r"""Compute the direct solar radiation.

    On page 19 of :cite:p:`cigre601`. Equation (10) states that the direct solar
    radiation on a surface normal to the solar beam at sea level, :math:`I_{B(0)}`, is given by

    .. math::

        N_s \frac{1280 \sin(H_s)}{\sin(H_s) + 0.314},

    where :math:`H_s` is the solar altitude.
    To correct for height above sea level, we use the Eq. 19 from Cigre 601,
    since no equation is provided in Cigre 207.

    Parameters
    ----------
    sin_solar_altitude:
        :math:`\sin\left(H_s\right)`. The sine of the solar altitude.
    height_above_sea_level:
        :math:`y~\left[\text{m}\right]`. The conductor's altitude.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`I_D~\left[\text{W}~\text{m}^{-2}\right]`. The direct solar radiation.
    """
    clearness_ratio = 1.0
    return cigre601.solar_heating.compute_direct_solar_radiation(
        sin_solar_altitude, clearness_ratio, height_above_sea_level
    )


def compute_diffuse_sky_radiation(
    direct_solar_radiation: WattPerSquareMeter,
    sin_solar_altitude: Unitless,
) -> WattPerSquareMeter:
    r"""Compute the diffuse radiation (light scattered in the atmosphere).

    On page 38 of :cite:p:`cigre207`.

    This equation differ from :cite:p:`cigre601`, however the difference is small, and the
    diffuse radiation is a small contributor to the overall solar radiation, so the total
    discrepancy between the models is small.

    Parameters
    ----------
    direct_solar_radiation:
        :math:`I_D~\left[\text{W}~\text{m}^{-2}\right]`. The direct solar radiation.
    sin_solar_altitude:
        :math:`\sin\left(H_s\right)`. The sine of the solar altitude.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`I_d~\left[\text{W}~\text{m}^{-2}\right]`.The diffuse solar radiation.
    """
    sin_H_s = sin_solar_altitude
    I_D = direct_solar_radiation
    return np.maximum(0, (570 - 0.47 * I_D)) * np.maximum(0, sin_H_s) ** 1.2


def compute_global_radiation_intensity(
    direct_solar_radiation: WattPerSquareMeter,
    diffuse_sky_radiation: WattPerSquareMeter,
    albedo: Unitless,
    sin_angle_of_sun_on_line: Unitless,
    sin_solar_altitude: Unitless,
) -> WattPerSquareMeter:
    r"""Compute the global radiation intensity experienced by the conductor.

    Equation (47) on page 38 of :cite:p:`cigre207` state that the global radiation intensity,
    :math:`I_T`, is given by

    .. math::

        I_T =
            I_D \left(\sin(\eta) + 0.5 F \pi \sin(H_s)\right) +
            0.5 I_d \pi \left(1 + F \right),

    where :math:`\eta` is the incidence angle of the sun on the line, :math:`H_s` is the solar
    altitude and :math:`F` is the ground albedo (amount of radiation diffusely reflected from the
    ground). The factor :math:`0.5 \pi` is due the assumption that the ground reflects light
    diffusely and uniformly in all directions, so the reflected energy is always directed normally
    to the line. In CIGRE207, it is also assumed that the diffuse radiation is uniformly directed,
    which leads to :math:`I_d \pi/2 (1 + F)` instead of  :math:`I_d (1 + 0.5 F \pi)`

    Parameters
    ----------
    direct_solar_radiation:
        :math:`I_D~\left[\text{W}~\text{m}^{-2}\right]`. The direct solar radiation.
    diffuse_sky_radiation:
        :math:`I_d~\left[\text{W}~\text{m}^{-2}\right]`.The diffuse solar radiation.
    albedo:
        :math:`F`. The ground albedo.
    sin_angle_of_sun_on_line:
        :math:`\sin\left(\eta\right)`. The sine of the angle of the sun on the line.
    sin_solar_altitude:
        :math:`\sin\left(H_s\right)`. The sine of the solar altitude.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`I_T~\left[\text{W}~\text{m}^{-2}\right]`. The global radiation intensity.

    Note
    ----
    The following values are given for the albedo in :cite:p:`cigre601`:

    .. list-table::
        :widths: 50 50
        :header-rows: 1

        * - Ground
          - Albedo
        * - Water (:math:`H_s > 30^\circ`)
          - 0.05
        * - Forest
          - 0.1
        * - Urban areas
          - 0.15
        * - Soil, grass and crops
          - 0.2
        * - Sand
          - 0.3
        * - Ice
          - 0.4-0.6
        * - Snow
          - 0.6-0.8
    """
    I_D = direct_solar_radiation
    I_d = diffuse_sky_radiation
    F = albedo
    sin_H_s = sin_solar_altitude
    sin_eta = sin_angle_of_sun_on_line
    F_pi_half = 0.5 * pi * F

    return I_D * (sin_eta + F_pi_half * sin_H_s) + I_d * pi / 2 * (1 + F)  # type: ignore



import importlib.metadata as importlib_metadata

__version__ = importlib_metadata.version(__name__)

from . import equations  # noqa
from .model import *  # noqa
from .solver import *  # noqa
from .types import *  # noqa
